/*
 *
 */
#include "isp_debug.h"

#include <linux/device.h>
#include <linux/mm.h>
#include <linux/module.h>

#include "isp.h"
#include "ispreg.h"
#include "ispccdc.h"
#include "ispresizer.h"
#include "ispccdc_vp.h"

#define MIN_IN_WIDTH			32
#define MIN_IN_HEIGHT			32
#define MAX_IN_WIDTH_MEMORY_MODE	4095
#define MAX_IN_WIDTH_ONTHEFLY_MODE_ES1	1280
#define MAX_IN_WIDTH_ONTHEFLY_MODE_ES2	4095
#define MAX_IN_HEIGHT			4095

/*
 * __resizer_get_format - helper function for getting resizer format
 * @res   : pointer to resizer private structure
 * @pad   : pad number
 * @fh    : V4L2 subdev file handle
 * @which : wanted subdev format
 * return zero
 */
static struct v4l2_mbus_framefmt *
__ccdc_vp_get_format(struct isp_ccdc_vp_device *ccdc_vp, struct v4l2_subdev_fh *fh,
		     unsigned int pad, enum v4l2_subdev_format_whence which)
{
	if (which == V4L2_SUBDEV_FORMAT_TRY)
		return v4l2_subdev_get_try_format(fh, pad);
	else
		return &ccdc_vp->formats[pad];
}

/* -----------------------------------------------------------------------------
 * V4L2 subdev operations
 */
#if PRINT_GRAPH_EN
void print_graph(struct isp_pipeline *pipe)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER, pipe: %p\n", pipe);

	struct isp_device *isp = pipe->output->isp;
	struct media_entity *entity;
	struct media_pad *pad;

	entity = &pipe->output->video.entity;

	while (1) {
		ISP_DEBUG_MSG("entity addr: %p\n", entity);
		if (entity == NULL)
			break;
		ISP_DEBUG_MSG("entity->name: %s\n", entity->name);
		ISP_DEBUG_MSG("entity->stream_count: %d\n", entity->stream_count);

		pad = &entity->pads[0];
		if (!(pad->flags & MEDIA_PAD_FLAG_INPUT))
			break;

		pad = media_entity_remote_source(pad);
		if (pad == NULL ||
			media_entity_type(pad->entity) !=
				MEDIA_ENTITY_TYPE_V4L2_SUBDEV)
			break;

		entity = pad->entity;

		ISP_DEBUG_MSG("entity addr: %p\n", entity);
		ISP_DEBUG_MSG("entity->name: %s\n", entity->name);
		ISP_DEBUG_MSG("entity->stream_count: %d\n", entity->stream_count);
	}
}
#endif

/*
 * ccdc_vp_set_stream - Enable/Disable streaming on resizer subdev
 * @sd: ISP resizer V4L2 subdev
 * @enable: 1 == Enable, 0 == Disable
 *
 * The resizer hardware can't be enabled without a memory buffer to write to.
 * As the s_stream operation is called in response to a STREAMON call without
 * any buffer queued yet, just update the state field and return immediately.
 * The resizer will be enabled in resizer_video_queue().
 */
static int ccdc_vp_set_stream(struct v4l2_subdev *sd, int enable)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER, enable = %d\n", enable);

	struct isp_ccdc_vp_device *ccdc_vp = v4l2_get_subdevdata(sd);
	struct isp_device *isp = to_isp_device(ccdc_vp);
	int ret = 0;

	if (ccdc_vp->state == ISP_PIPELINE_STREAM_STOPPED) {
		if (enable == ISP_PIPELINE_STREAM_STOPPED)
			return 0;

		ccdc_vp->error = 0;
	}

	switch (enable) {
	case ISP_PIPELINE_STREAM_CONTINUOUS:
		ISP_CCDC_VP_DEBUG_MSG("case ISP_PIPELINE_STREAM_CONTINUOUS:\n");

		ret = ispccdc_to_res_en(isp, enable);

		break;
	case ISP_PIPELINE_STREAM_SINGLESHOT:
		ISP_CCDC_VP_DEBUG_MSG("case ISP_PIPELINE_STREAM_SINGLESHOT:\n");

		break;
	case ISP_PIPELINE_STREAM_STOPPED:
		ISP_CCDC_VP_DEBUG_MSG("case ISP_PIPELINE_STREAM_STOPPED:\n");

		ret = ispccdc_to_res_en(isp, enable);

		break;
	}

	ccdc_vp->state = enable;
	return ret;
}

/* resizer pixel formats */
static const unsigned int ccdc_vp_fmts[] = {
	V4L2_MBUS_FMT_UYVY8_1X16,
	V4L2_MBUS_FMT_YUYV8_1X16,
	V4L2_MBUS_FMT_UYVY8_2X8,
	V4L2_MBUS_FMT_VYUY8_2X8,
	V4L2_MBUS_FMT_YUYV8_2X8,
	V4L2_MBUS_FMT_YVYU8_2X8,
};

static unsigned int resizer_max_in_width(struct isp_ccdc_vp_device *ccdc_vp)
{
	struct isp_device *isp = to_isp_device(ccdc_vp);

	if (isp->revision == ISP_REVISION_1_0)
		return MAX_IN_WIDTH_ONTHEFLY_MODE_ES1;
	else
		return MAX_IN_WIDTH_ONTHEFLY_MODE_ES2;
}

/*
 * ccdc_vp_try_format - Handle try format by pad subdev method
 * @res   : ISP resizer device
 * @fh    : V4L2 subdev file handle
 * @pad   : pad num
 * @fmt   : pointer to v4l2 format structure
 * @which : wanted subdev format
 */
static void ccdc_vp_try_format(struct isp_ccdc_vp_device *ccdc_vp,
			       struct v4l2_subdev_fh *fh, unsigned int pad,
			       struct v4l2_mbus_framefmt *fmt,
			       enum v4l2_subdev_format_whence which)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER. pad = %d, which = %d\n", pad, which);

	struct v4l2_mbus_framefmt *format;

	switch (fmt->code) {
	case V4L2_MBUS_FMT_YUYV8_1X16:
	case V4L2_MBUS_FMT_UYVY8_1X16:
	case V4L2_MBUS_FMT_UYVY8_2X8:
	case V4L2_MBUS_FMT_VYUY8_2X8:
	case V4L2_MBUS_FMT_YUYV8_2X8:
	case V4L2_MBUS_FMT_YVYU8_2X8:
		break;
	default:
		fmt->code = V4L2_MBUS_FMT_YUYV8_1X16;
	}

	switch (pad) {
	case CCDC_VP_PAD_SOURCE:
	default:
		format = __ccdc_vp_get_format(ccdc_vp, fh, CCDC_VP_PAD_SOURCE, which);

		break;
	}

	fmt->colorspace = V4L2_COLORSPACE_JPEG;
	fmt->field = V4L2_FIELD_NONE;
}

/*
 * ccdc_vp_enum_mbus_code - Handle pixel format enumeration
 * @sd     : pointer to v4l2 subdev structure
 * @fh     : V4L2 subdev file handle
 * @code   : pointer to v4l2_subdev_mbus_code_enum structure
 * return -EINVAL or zero on success
 */
static int ccdc_vp_enum_mbus_code(struct v4l2_subdev *sd,
				  struct v4l2_subdev_fh *fh,
				  struct v4l2_subdev_mbus_code_enum *code)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	struct isp_ccdc_vp_device *ccdc_vp = v4l2_get_subdevdata(sd);
	struct v4l2_mbus_framefmt *format;

	if (code->pad == CCDC_VP_PAD_SOURCE) {
		if (code->index >= ARRAY_SIZE(ccdc_vp_fmts))
			return -EINVAL;

		code->code = ccdc_vp_fmts[code->index];
	} else {
		if (code->index != 0)
			return -EINVAL;

		format = __ccdc_vp_get_format(ccdc_vp, fh, CCDC_VP_PAD_SOURCE,
					      V4L2_SUBDEV_FORMAT_TRY);
		code->code = format->code;
	}

	return 0;
}

static int ccdc_vp_enum_frame_size(struct v4l2_subdev *sd,
				   struct v4l2_subdev_fh *fh,
				   struct v4l2_subdev_frame_size_enum *fse)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	struct isp_ccdc_vp_device *ccdc_vp = v4l2_get_subdevdata(sd);
	struct v4l2_mbus_framefmt format;

	if (fse->index != 0)
		return -EINVAL;

	format.code = fse->code;
	format.width = 1;
	format.height = 1;
	ccdc_vp_try_format(ccdc_vp, fh, fse->pad, &format, V4L2_SUBDEV_FORMAT_TRY);
	fse->min_width = format.width;
	fse->min_height = format.height;

	if (format.code != fse->code)
		return -EINVAL;

	format.code = fse->code;
	format.width = -1;
	format.height = -1;
	ccdc_vp_try_format(ccdc_vp, fh, fse->pad, &format, V4L2_SUBDEV_FORMAT_TRY);
	fse->max_width = format.width;
	fse->max_height = format.height;

	return 0;
}

/*
 * resizer_get_format - Handle get format by pads subdev method
 * @sd    : pointer to v4l2 subdev structure
 * @fh    : V4L2 subdev file handle
 * @fmt   : pointer to v4l2 subdev format structure
 * return -EINVAL or zero on sucess
 */
static int ccdc_vp_get_format(struct v4l2_subdev *sd, struct v4l2_subdev_fh *fh,
			      struct v4l2_subdev_format *fmt)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	struct isp_ccdc_vp_device *ccdc_vp = v4l2_get_subdevdata(sd);
	struct v4l2_mbus_framefmt *format;

	format = __ccdc_vp_get_format(ccdc_vp, fh, fmt->pad, fmt->which);
	if (format == NULL)
		return -EINVAL;

	fmt->format = *format;

	ISP_CCDC_VP_DEBUG_MSG("fmt->pad: %d\n", fmt->pad);
	ISP_CCDC_VP_DEBUG_MSG("format->code: 0x%04x\n", format->code);
	ISP_CCDC_VP_DEBUG_MSG("format->width: %d\n", format->width);
	ISP_CCDC_VP_DEBUG_MSG("format->height: %d\n", format->height);

	return 0;
}

/*
 * ccdc_vp_set_format - Handle set format by pads subdev method
 * @sd    : pointer to v4l2 subdev structure
 * @fh    : V4L2 subdev file handle
 * @fmt   : pointer to v4l2 subdev format structure
 * return -EINVAL or zero on success
 */
static int ccdc_vp_set_format(struct v4l2_subdev *sd, struct v4l2_subdev_fh *fh,
			      struct v4l2_subdev_format *fmt)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	struct isp_ccdc_vp_device *ccdc_vp = v4l2_get_subdevdata(sd);
	struct isp_device *isp = to_isp_device(ccdc_vp);
	struct v4l2_mbus_framefmt *format;
	struct v4l2_mbus_framefmt *ccdc_format;
	struct v4l2_subdev_format ccdc_fmt;

	format = __ccdc_vp_get_format(ccdc_vp, fh, fmt->pad, fmt->which);
	if (format == NULL)
		return -EINVAL;

	ISP_CCDC_VP_DEBUG_MSG("isp->isp_ccdc.formats[CCDC_PAD_SOURCE_OF].code = 0x%x\n" \
				 "\t\tisp->isp_ccdc.formats[CCDC_PAD_SOURCE_OF].width = %d\n" \
				 "\t\tisp->isp_ccdc.formats[CCDC_PAD_SOURCE_OF].height = %d\n", \
				 isp->isp_ccdc.formats[CCDC_PAD_SOURCE_OF].code, \
				 isp->isp_ccdc.formats[CCDC_PAD_SOURCE_OF].width, \
				 isp->isp_ccdc.formats[CCDC_PAD_SOURCE_OF].height);

	ccdc_format = &isp->isp_ccdc.formats[CCDC_PAD_SOURCE_OF];
	ccdc_fmt.pad = fmt->pad;
	ccdc_fmt.which = fmt->which;
	ccdc_fmt.format = *ccdc_format;

	ccdc_vp_try_format(ccdc_vp, fh, ccdc_fmt.pad, &ccdc_fmt.format, ccdc_fmt.which);

	ISP_CCDC_VP_DEBUG_MSG("ccdc_fmt.format.code: 0x%04x, fmt->format.code: 0x%04x\n", ccdc_fmt.format.code, fmt->format.code);
	if (ccdc_fmt.format.code != fmt->format.code)
		return -EINVAL;

	*format = ccdc_fmt.format;

	fmt->format = *format;

	ISP_CCDC_VP_DEBUG_MSG("fmt->pad: %d\n", fmt->pad);
	ISP_CCDC_VP_DEBUG_MSG("fmt->format.code: 0x%04x\n", fmt->format.code);
	ISP_CCDC_VP_DEBUG_MSG("fmt->format.width: %d\n", fmt->format.width);
	ISP_CCDC_VP_DEBUG_MSG("fmt->format.height: %d\n", fmt->format.height);

	// update resiser input
	struct isp_res_device *res = &isp->isp_res;

	ISP_CCDC_VP_DEBUG_MSG("res->state = %d\n", res->state);
	if(res->state) {
		ispresizer_set_crop_params(res, &res->formats[RESZ_PAD_SINK], &res->formats[RESZ_PAD_SOURCE]);
//		ispresizer_set_input_size(res, res->crop.active.width,
//						res->crop.active.height);
	}


	return 0;
}

/*
 * ccdc_vp_init_formats - Initialize formats on all pads
 * @sd: ISP ccdc_vp V4L2 subdevice
 * @fh: V4L2 subdev file handle
 *
 * Initialize all pad formats with default values. If fh is not NULL, try
 * formats are initialized on the file handle. Otherwise active formats are
 * initialized on the device.
 */
static int ccdc_vp_init_formats(struct v4l2_subdev *sd,
				struct v4l2_subdev_fh *fh)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	struct isp_ccdc_vp_device *ccdc_vp = v4l2_get_subdevdata(sd);

	ccdc_vp->formats[0].code = V4L2_MBUS_FMT_UYVY8_2X8;
	ccdc_vp->formats[0].width = 4096;
	ccdc_vp->formats[0].height = 4096;
	ccdc_vp->formats[0].colorspace = V4L2_COLORSPACE_JPEG;
	ccdc_vp->formats[0].field = V4L2_FIELD_NONE;

	return 0;
}

static int ccdc_vp_open(struct v4l2_subdev *sd,
				struct v4l2_subdev_fh *fh)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	return 0;
}

/* subdev core operations */
static const struct v4l2_subdev_core_ops ccdc_vp_v4l2_core_ops = {
//	.queryctrl = v4l2_subdev_queryctrl,
//	.querymenu = v4l2_subdev_querymenu,
//	.g_ctrl = v4l2_subdev_g_ctrl,
//	.s_ctrl = v4l2_subdev_s_ctrl,
//	.g_ext_ctrls = v4l2_subdev_g_ext_ctrls,
//	.try_ext_ctrls = v4l2_subdev_try_ext_ctrls,
//	.s_ext_ctrls = v4l2_subdev_s_ext_ctrls,
};

/* subdev file operations */
static const struct v4l2_subdev_file_ops ccdc_vp_v4l2_file_ops = {
//	.open = ccdc_vp_init_formats,
	.open = ccdc_vp_open,
};

/* subdev video operations */
static const struct v4l2_subdev_video_ops ccdc_vp_v4l2_video_ops = {
	.s_stream = ccdc_vp_set_stream,
};

/* subdev pad operations */
static const struct v4l2_subdev_pad_ops ccdc_vp_v4l2_pad_ops = {
	.enum_mbus_code = ccdc_vp_enum_mbus_code,
	.enum_frame_size = ccdc_vp_enum_frame_size,
	.get_fmt = ccdc_vp_get_format,
	.set_fmt = ccdc_vp_set_format,
//	.get_crop = resizer_g_crop,
//	.set_crop = resizer_s_crop,
};

/* subdev operations */
static const struct v4l2_subdev_ops ccdc_vp_v4l2_ops = {
	.core = &ccdc_vp_v4l2_core_ops,
	.file = &ccdc_vp_v4l2_file_ops,
	.video = &ccdc_vp_v4l2_video_ops,
	.pad = &ccdc_vp_v4l2_pad_ops,
};


/* -----------------------------------------------------------------------------
 * Media entity operations
 */

/*
 * ccdc_vpr_init_entities - Initialize ccdc_vp subdev and media entity.
 * @res : Pointer to ccdc_vp device structure
 * return -ENOMEM or zero on success
 */
static int ispccdc_vp_init_entities(struct isp_ccdc_vp_device *ccdc_vp)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	struct v4l2_subdev *sd = &ccdc_vp->subdev;
	struct media_pad *pads = ccdc_vp->pads;
	struct media_entity *me = &sd->entity;
	int ret;

	v4l2_subdev_init(sd, &ccdc_vp_v4l2_ops);
	strlcpy(sd->name, "OMAP3 ISP CCDC VP", sizeof(sd->name));
	sd->grp_id = 1 << 16;	/* group ID for isp subdevs */
	v4l2_set_subdevdata(sd, ccdc_vp);
	sd->flags |= V4L2_SUBDEV_FL_HAS_DEVNODE;

	v4l2_ctrl_handler_init(&ccdc_vp->ctrls, 1);
	sd->ctrl_handler = &ccdc_vp->ctrls;

	pads[CCDC_VP_PAD_SOURCE].flags = MEDIA_PAD_FLAG_OUTPUT;

	ret = media_entity_init(me, CCDC_VP_PADS_NUM, pads, 0);
	if (ret < 0)
		return ret;

	ccdc_vp_init_formats(sd, NULL);

	return 0;
}

void isp_ccdc_vp_unregister_entities(struct isp_ccdc_vp_device *ccdc_vp)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	media_entity_cleanup(&ccdc_vp->subdev.entity);

	v4l2_device_unregister_subdev(&ccdc_vp->subdev);
	v4l2_ctrl_handler_free(&ccdc_vp->ctrls);
}

int isp_ccdc_vp_register_entities(struct isp_ccdc_vp_device *ccdc_vp,
				  struct v4l2_device *vdev)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	int ret;

	/* Register the subdev and video nodes. */
	ret = v4l2_device_register_subdev(vdev, &ccdc_vp->subdev);
	if (ret < 0)
		goto error;

	return 0;

error:
	isp_ccdc_vp_unregister_entities(ccdc_vp);
	return ret;
}

/* -----------------------------------------------------------------------------
 * ISP ccdc_vp initialization and cleanup
 */

void isp_ccdc_vp_cleanup(struct isp_device *isp)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");
}

/*
 * isp_ccdc_vp_init - CCDC_VP initialization.
 * @isp : Pointer to ISP device
 * return -ENOMEM or zero on success
 */
int isp_ccdc_vp_init(struct isp_device *isp)
{
	ISP_CCDC_VP_DEBUG_MSG("ENTER\n");

	struct isp_ccdc_vp_device *ccdc_vp = &isp->isp_ccdc_vp;
	int ret;

	ret = ispccdc_vp_init_entities(ccdc_vp);
	if (ret < 0)
		goto out;

out:
	if (ret)
		isp_ccdc_vp_cleanup(isp);

	return ret;
}
